﻿/*
 * Copyright 2014 UOC
 */

/**
 *
 * @file       main.cpp
 * @author     Albert Creixell
 * @version    v0.1
 * @date       Oct, 2014
 * @brief
 * @ingroup
 *
 */

/*================================ include ==================================*/

#include "string.h"

#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

#include "openmote-cc2538.h"

#include "Callback.h"
#include "Serial.h"

/*================================ define ===================================*/

#define PAYLOAD_LENGTH                      ( 125 )
#define EUI48_LENGTH                        ( 6 )


/*================================ typedef ==================================*/

/*=============================== prototypes ================================*/

extern "C" void vApplicationTickHook(void);
extern "C" void vApplicationIdleHook(void);

static void rtc_callback(void);
static PlainCallback rtcCallback(rtc_callback);


static void iniciaranuraCord(void);
static void iniciaranuraEsc(void);
static void sincronitza(void);
static void escriu(void);
static void llegeix(void);
static void incrementaASN(void);
static void incrementaRanura(void);
static void rxInit(void);
static void rxDone(void);
static void txInit(void);
static void txDone(void);

/*=============================== variables =================================*/

static xSemaphoreHandle rxSemaphore;
static xSemaphoreHandle txSemaphore;

static PlainCallback rxInitCallback(&rxInit);
static PlainCallback rxDoneCallback(&rxDone);
static PlainCallback txInitCallback(&txInit);
static PlainCallback txDoneCallback(&txDone);

static uint8_t radio_escriptura[PAYLOAD_LENGTH];
static uint8_t* radio_ptr = radio_escriptura;
static uint8_t  radio_len = sizeof(radio_escriptura);
static RadioResult result;

static uint8_t radio_lectura[PAYLOAD_LENGTH];
static uint8_t* radio_ptr2 = radio_lectura;
static uint8_t  radio_len2 = sizeof(radio_lectura);
static int8_t rssi;
static uint8_t lqi;
static uint8_t crc;

static uint8_t uart_buffer[PAYLOAD_LENGTH];
static uint8_t* uart_ptr = radio_escriptura;
static uint8_t  uart_len = sizeof(radio_escriptura);

static Serial serial(uart);

/*================================= TSCH ==================================*/
static uint8_t canal=26;
static uint8_t sincronitzat=0;
static bool assignat=false;
static uint8_t ASN[4];
static uint8_t numranures=21;
static uint8_t ordrecom[21][3];
static uint8_t cont_ranura=0;
static uint8_t estat=0;
static uint8_t tipus=0;
static uint8_t ident=1;
static bool coordinador=true;
static uint32_t tic1;
static uint32_t tic2;
static uint32_t tic3;
static uint32_t tic4;
static int32_t retras;
static int32_t correccio;
static bool registra;
static bool llegit=false;
static bool verif=false;

/*================================= public ==================================*/

int main (void)
{
    // Set the TPS62730 in bypass mode (Vin = 3.3V, Iq < 1 uA)
    //tps62730.setBypass();
    
    // Enable erasing the Flash with the user button
    //board.enableFlashErase();

    // Enable the IEEE 802.15.4 radio
    radio.setTxCallbacks(&txInitCallback, &txDoneCallback);
    radio.setRxCallbacks(&rxInitCallback, &rxDoneCallback);
    radio.enable();
    radio.enableInterrupts();
    //radio.setChannel(RADIO_CHANNEL);

    //Inicialitza el vector de comunicacions
    ASN[0]=0;ASN[1]=0;ASN[2]=0;ASN[3]=0;ASN[4]=0;
    uint8_t contador;
    for(contador=0;contador<=numranures;contador++){
	ordrecom[contador][0]=0;ordrecom[contador][1]=0;ordrecom[contador][2]=0;ordrecom[contador][3]=0;
    }

    // Activa el port sèrie del coordinador per poder veure dades
    //if (coordinador==true){
	uart.enable(UART_BAUDRATE, UART_CONFIG, UART_INT_MODE);
	serial.init();
    //}


    //Initialize Rtc
    rtc.init();
    rtc.setCallback(&rtcCallback);
    rtc.enableInterrupt();

    // Start Rtc
    rtc.start(327);

    // Enable interrupts
    board.enableInterrupts();

    vTaskStartScheduler();
    
    // Forever
    /*while (true)
    {
        // Sleep
        board.setSleepMode(SleepMode::SleepMode_2);
        board.sleep();
    }*/
}

/*================================ private ==================================*/

static void rtc_callback(void) //El callback de la interrupció del RTC s'utilitza per controlar tot el procés
{
       
    if (coordinador==true){ //Codi que s'executa si es coordinador
	sincronitzat=1;	//El coordinador és el master i la seva base de temps sempre és la correcte
	if (estat==0){iniciaranuraCord();estat=1;} //Primer estat de la ranura, on s'inicialitza que ha de fer
	if (estat==1 && tipus==0){incrementaRanura();incrementaASN();estat=0;rtc.start(393);} //Tipus=0 no fa res espera 12ms i inicia la ranura següent
	if (estat==1 && tipus==1){estat=2;} //Tipus=1 implica enviar informació
	if (estat==1 && tipus==2){estat=10;} //Tipus=2 implica llegir informació
	if (estat>1 && estat<10){escriu();} //estats d'escriptura
	if (estat>9 && estat<20){llegeix();} //estats de lectura
    }
    else{	
	if (sincronitzat==1){	//Si es esclau i està sincronitzat executa aquest codi
		led_red.off();
		//if(ordrecom[3][0]!=0){led_orange.on();}
		if (estat==0){iniciaranuraEsc();estat=1;}
		if (estat==1 && tipus==0){incrementaRanura();incrementaASN();estat=0;rtc.start(393);} //Tipus=0 no fa res espera 12ms i inicia la ranura
		if (estat==1 && tipus==1){estat=2;}
		if (estat==1 && tipus==2){estat=10;}
		if (estat>1 && estat<10){escriu();}
		if (estat>9 && estat<20){llegeix();}
	}
	else{
		if (estat<20){estat=20;} //Estat de sincronització
		sincronitza();
	}
    }
}

static void iniciaranuraCord(void)
{
    if (cont_ranura==0){
	radio_escriptura[0]=255;
	radio_escriptura[1]=ASN[0];radio_escriptura[2]=ASN[1];radio_escriptura[3]=ASN[2];
	radio_escriptura[4]=ASN[3];radio_escriptura[5]=ASN[4];
	tipus=1;canal=26;radio.setChannel(canal);radio_len=6; //Envia l'ASN a tots els esclaus perque sincronitzin
	
    }
    if (cont_ranura==1){
	tipus=2;canal=26;radio.setChannel(canal); //Ranura on escolta algun esclau i li assigna ranura en el ACK 
    }
    if (cont_ranura>=2){
	tipus=0;//Ranures on escriu, llegeix o no fa res segosn les ranures assignades
	canal=26;radio.setChannel(canal);	
	if (ordrecom[cont_ranura][0]==ident){tipus=1;}
	if (ordrecom[cont_ranura][1]==ident){tipus=2;}
	
    }
    
}

static void iniciaranuraEsc(void)
{
    if (cont_ranura==0){
	tipus=0;canal=26;radio.setChannel(canal); //Un cop sincronitzat, l'esclau ignora la ranura 0
    }
    if (cont_ranura==1){
	if (assignat==false){			
		tipus=1;
		canal=26;radio.setChannel(canal); //Ranura demanarà si no enté ranura per rebre i enviar
		radio_escriptura[0]=10;
		radio_escriptura[1]=20;radio_escriptura[2]=ident;
		radio_len=3;
	}
	else {tipus=0;}
    }
    if (cont_ranura>=2){
	tipus=0;//Ranures on escriu, llegeix o no fa res segosn les ranures assignades
	canal=26;radio.setChannel(canal);	
	if (ordrecom[cont_ranura][0]==ident){tipus=1;}
	if (ordrecom[cont_ranura][1]==ident){tipus=2;}
    }
    
}

static void sincronitza(void)
{
    switch (estat){
	case 20:
		radio.setChannel(canal);
		led_red.on();		
		radio.on();
		radio.receive();// Posa la radio a escoltar permanentment. En futur s'implementarà duty cycle
		estat=21;
		
    }	
}

static void escriu(void)
{
    switch (estat){
	case 2:		
		radio.on();		
		radio_ptr = radio_escriptura;
	        result = radio.loadPacket(radio_ptr, radio_len);		
    		estat=3;
		rtc.start(98); //Carrega paquet a la radio i espera 3ms per enviar
	case 3:
		//if (result == RadioResult_Success && cont_ranura==1 && verif==false){led_orange.on();verif=true;}else{led_orange.off();verif=false;}	
		uart_ptr = uart_buffer; //això ho faig per verificar que els buffer d'escriptura estan amb el missatge que vull i ho estan. Verificat
                memcpy(uart_ptr, radio_ptr, radio_len);
                uart_ptr += radio_len;
                uart_len = radio_len;              
		uart.writeByte(uart_buffer, uart_len);		
		radio.transmit();tic1=rtc.read();estat=4;rtc.start(163); //Es carrega 163 (5ms) al rtc per controlar el temps que triga en enviar.
        case 4:
		llegit=false;		
		radio.on();
		radio.receive();// Put the radio transceiver in receive mode
		tic3=rtc.read();
		estat=5;
		rtc.start(64);//Es carrega 64 (2ms) al rtc per controlar el temps que triga en enviar.
	case 5:	
		radio_ptr2 = radio_lectura;
            	radio_len2 = sizeof(radio_lectura);
            	result = radio.getPacket(radio_ptr2, &radio_len2, &rssi, &lqi, &crc);
		radio.off();
		if (llegit==true && coordinador==false && cont_ranura==1 && assignat==false){
			int index=0;			
			if (radio_lectura[0]==10 && radio_lectura[1]==25){
				index=radio_lectura[4];				
				ordrecom[index][0]=ident;
				ordrecom[index][1]=1;
				ordrecom[index+1][0]=1;
				ordrecom[index+1][1]=ident;
				assignat=true;
			}
		}
		incrementaRanura();incrementaASN();estat=0;rtc.start((163+64));//rtc.start((163-(tic2-tic1)+64-(tic4-tic3)));
    }
}

static void llegeix(void)
{
    switch (estat){
	case 9:
		llegit=false;		
		radio.on();
		radio.receive();// Put the radio transceiver in receive mode
		tic1=rtc.read();		
		rtc.start(295); //Inicialitzem el rellotge per controlar el retràs. Idealment les com començen a tick 98
		estat=10;
        case 10:
		estat=11;
        case 11:
		radio_ptr2 = radio_lectura;
            	radio_len2 = sizeof(radio_lectura);
            	result = radio.getPacket(radio_ptr2, &radio_len2, &rssi, &lqi, &crc);		
		if (llegit==true){					
			uart_ptr = uart_buffer;
                	memcpy(uart_ptr, radio_ptr2, radio_len2);
                	uart_ptr += radio_len2;
                	uart_len = radio_len2;              
			uart.writeByte(uart_buffer, uart_len);
			radio.on();
			if (coordinador==true){
				//Si coordinador, els bytes 2 i 3 indiquen el retràs detectat en el ACK				
				radio_escriptura[0]=30;radio_escriptura[1]=60;
				radio_escriptura[2]=retras;radio_escriptura[3]=retras >> 8;radio_escriptura[4]=0;
				if (cont_ranura==1){ //Si es la ranura 1 i coordinador envia en el ACK la ranura a utilitzar
					uint8_t contador;				
					registra=false;
					contador=3;
					if (radio_lectura[0]==10 && radio_lectura[1]==20){					
						while(registra=false){
							if (ordrecom[contador][0]==0 && ordrecom[contador][1]==0){
								//S'asigna ranura en parelles. Primer esclau envia i a la següent ranura escolta al coordinador
								ordrecom[contador][0]=radio_lectura[2];
								ordrecom[contador][1]=ident;
								ordrecom[contador+1][0]=ident;
								ordrecom[contador+1][1]=radio_lectura[2];
								registra=true;
								radio_escriptura[0]=10;radio_escriptura[1]=25;
								radio_escriptura[4]=contador;
							}
							else{
								if (contador>=numranures){registra=true;}
								else{contador=contador+2;}
							}
						}
					}
				}
			}
			else{
				//Si no es coordinador en el ACK no s'indica correcció de retràs				
				radio_escriptura[0]=30;radio_escriptura[1]=60;
				radio_escriptura[2]=0;radio_escriptura[3]=0;
				radio_escriptura[4]=0;
			}
			radio_ptr = radio_escriptura;
		        result = radio.loadPacket(radio_ptr, radio_len);		
	    		estat=12;
			rtc.start(32); //Carrega paquet a la radio i espera 1ms per enviar
		}
		else{
			tic4=0;tic3=0;estat=13;rtc.start(96);
		}
	case 12:
		radio.transmit();tic3=rtc.read();estat=13;rtc.start(64); //Envia el ACK
	case 13:
		incrementaRanura();incrementaASN();estat=0; //Finalitza lectura i si es esclau aplica correcció de temps
		if (coordinador==true){		
			rtc.start((295+64));//rtc.start((295-(tic2-tic1)+64-(tic4-tic3)));
		}
		else{
			correccio=0;
			if (llegit==true && radio_lectura[0]==30 && radio_lectura[1]==60){			
				correccio=radio_lectura[3]<<8;
				correccio=correccio+radio_lectura[2];
			}
			rtc.start((295+64));//rtc.start((295-(tic2-tic1)+64-(tic4-tic3))+correccio);
		}
		
    }
}
    


static void incrementaASN(void)
{
    if (ASN[0]==255){ASN[0]=0;ASN[1]++;}
    else {ASN[0]++;}
    
    if (ASN[1]==255){ASN[1]=0;ASN[2]++;}
    if (ASN[2]==255){ASN[2]=0;ASN[3]++;}  
    if (ASN[3]==255){ASN[3]=0;ASN[4]++;}
    if (ASN[4]==255){ASN[4]=0;}   
}

static void incrementaRanura(void)
{
    cont_ranura++;
    if (cont_ranura>numranures){
	cont_ranura=0;
	//if (verif==false){led_orange.on();verif=true;}else{led_orange.off();verif=false;}
    } 
}


static void rxInit(void)
{
    // Turn on the radio LED as the radio is now receiving a packet
    if (estat==10){tic2=rtc.read();retras=(tic2-tic1)-98;}
    led_green.on();
}

static void rxDone(void)
{
    static BaseType_t xHigherPriorityTaskWoken;
    xHigherPriorityTaskWoken = pdFALSE;
    xSemaphoreGiveFromISR(rxSemaphore, &xHigherPriorityTaskWoken);
    portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
    led_green.off();
    llegit=true;
    if (estat==4){tic3=rtc.read();rtc.stop();estat=5;rtc.start(1);}
    if (estat==10){tic2=rtc.read();estat=11;radio.off();rtc.start(1);}
    if (estat==21){
	radio_ptr2 = radio_lectura;
        radio_len2 = sizeof(radio_lectura);
        result = radio.getPacket(radio_ptr2, &radio_len2, &rssi, &lqi, &crc);		
	if (radio_lectura[0]==255){
		ASN[0]=radio_lectura[1];ASN[1]=radio_lectura[2];
		ASN[2]=radio_lectura[3];ASN[3]=radio_lectura[4];
		ASN[4]=radio_lectura[5];
		sincronitzat=1;cont_ranura=0;
		incrementaRanura();incrementaASN();estat=0;
		radio.off();
		rtc.start(262); //Aproximadament la pròxima ranura serà a 8ms. L'ajust fi es farà en la ranura 1 amb l'ACK	
	}
    }
}

static void txInit(void)
{
    // Turn on the radio LED as the packet is now transmitting
	led_yellow.on();    
}

static void txDone(void)
{
    static BaseType_t xHigherPriorityTaskWoken;
    xHigherPriorityTaskWoken = pdFALSE;
    xSemaphoreGiveFromISR(txSemaphore, &xHigherPriorityTaskWoken);
    portYIELD_FROM_ISR(xHigherPriorityTaskWoken);	
    led_yellow.off();
    if (estat==3){tic2=rtc.read();rtc.stop();radio.off();estat=4;rtc.start(64);}
    if (estat==12){tic4=rtc.read();rtc.stop();radio.off();estat=13;rtc.start(1);}

}
